from typing import Any

from vstools import CustomIndexError, CustomValueError, DependencyNotFoundError, FieldBased, core, vs

from .types import ValidMatchT

__all__ = [
    'FieldMatches',
]


class FieldMatches(list[str]):
    """Class for holding field matches."""

    def __init__(self, matches: list[str] | None = None) -> None:
        super().__init__(matches or [])

    def __contains__(self, match: str | int | Any) -> bool:
        if isinstance(match, str):
            return any(m == match for m in self)

        if isinstance(match, int):
            return any(m == match for m in self)

        return match in self

    def __init_subclass__(cls) -> None:
        """Create match properties for the class."""

        for match in ValidMatchT.__args__:  # type: ignore

            def create_match_property(match: ValidMatchT) -> property:
                """Create a match property for the class."""

                def getter(self: FieldMatches) -> list[str]:
                    """Get all frames with a specific match."""

                    return [f for f in self if f.match == match]

                return property(getter, doc=f"Get all frames with a '{match}' match.")

            setattr(cls, f'{match}_matches', create_match_property(match))

    def __str__(self) -> str:
        return ', '.join(str(match) for match in self)

    @classmethod
    def wob_json_key(cls) -> str:
        """The JSON key for matches."""

        return 'matches'

    @property
    def fieldhint_string(self) -> str:
        """Get a string representation of the matches to pass to FieldHint."""

        return ''.join(self)

    def get_match_at_frame(self, frame: int) -> str:
        """Get the match value for a given frame index."""

        try:
            return self[frame]
        except IndexError:
            raise CustomIndexError(f'Frame {frame} is out of bounds (0-{len(self)})!', self.get_match_at_frame)

    def set_orphans_to_combed_matches(self, orphans: Any) -> None:
        """Set the matches of orphan frames to 'c'."""

        from .orphans import OrphanFrames

        if not isinstance(orphans, OrphanFrames):
            raise CustomValueError('Orphans must be an OrphanFrames instance!', orphans)

        for orphan in orphans:
            self[orphan.frame] = 'c'

    def apply(self, clip: vs.VideoNode) -> vs.VideoNode:
        """Apply the matches to the clip."""

        if not hasattr(core, 'fh'):
            raise DependencyNotFoundError(self.apply, 'FieldHint')

        fh = clip.fh.FieldHint(tff=FieldBased.from_video(clip).is_tff, matches=self.fieldhint_string)

        match_clips = dict[str, vs.VideoNode]()

        for match in set(self):
            match_clips[match] = fh.std.SetFrameProps(WobblyMatch=match)

        return fh.std.FrameEval(lambda n: match_clips[self[n]])



from typing import Any

from vsdeinterlace import FixInterlacedFades
from vstools import CustomTypeError, replace_ranges, vs

from ...data.parse import WobblyParser
from ...types import FilteringPositionEnum
from .abstract import AbstractProcessingStrategy

__all__ = [
    'AdaptiveFixInterlacedFadesStrategy',
]


class _FrameRangeGrouper:
    """Helper class for grouping frame numbers into ranges."""

    def group_frames_into_ranges(
        self, clip: vs.VideoNode, frame_nums: list[int], max_gap: int = 3
    ) -> list[tuple[int, int]]:
        """
        Group frame numbers into ranges if they are close enough together.

        :param clip:        Clip used to determine valid frame range
        :param frame_nums:  List of frame numbers to group
        :param max_gap:     Maximum gap between frames to consider them part of the same range

        :return:           List of (start, end) frame ranges
        """

        frame_nums = [
            x.frame if not isinstance(x, int) and hasattr(x, 'frame') and isinstance(x.frame, int) else x
            for x in frame_nums
        ]

        if invalid_frames := [x for x in frame_nums if not isinstance(x, int)]:
            raise CustomTypeError(
                'All frame numbers must be integers or have an integer frame attribute!',
                self.group_frames_into_ranges,
                invalid_frames,
            )

        frame_nums = sorted(frame_nums)
        last_frame = clip.num_frames - 1

        ranges = []
        start = end = frame_nums[0]

        for curr in frame_nums[1:]:
            if curr > last_frame:
                break

            if curr - end > max_gap:
                ranges.append((start, end))
                start = curr

            end = curr

        if end <= last_frame:
            ranges.append((start, end))

        return ranges


class AverageFixInterlacedFadesStrategy(AbstractProcessingStrategy):
    """Strategy for fixing interlaced fades using the Average method."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._frame_grouper = _FrameRangeGrouper()

    def apply(self, clip: vs.VideoNode, wobbly_parsed: WobblyParser) -> vs.VideoNode:
        """
        Use FixInterlacedFades.Average to fix interlaced fades.

        :param clip:            The clip to process.
        :param wobbly_parsed:   The parsed wobbly file. See the `WobblyParser` class for more information,
                                including all the data that is available.
        :param kwargs:          Additional keyword arguments.

        :return:                Clip with the processing applied to the selected frames.
        """

        frame_groups = self._frame_grouper.group_frames_into_ranges(clip, wobbly_parsed.interlaced_fades, 5)

        return replace_ranges(clip, FixInterlacedFades.Average(clip), frame_groups)

    @property
    def position(self) -> FilteringPositionEnum:
        """When to perform filtering."""

        return FilteringPositionEnum.PRE_DECIMATE


class AdaptiveFixInterlacedFadesStrategy(AbstractProcessingStrategy):
    """Strategy for fixing interlaced fades adaptively."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._frame_grouper = _FrameRangeGrouper()

    def apply(self, clip: vs.VideoNode, wobbly_parsed: WobblyParser, **kwargs: Any) -> vs.VideoNode:
        """
        Use FixInterlacedFades to fix interlaced fades, and apply it adaptively.

        This works by first combining all (nearly) adjacent frames into a list, and then comparing surrounding frames
        to determine if it's a fade from black or white. If it's unsure, it'll average instead.

        :param clip:            The clip to process.
        :param wobbly_parsed:   The parsed wobbly file. See the `WobblyParser` class for more information,
                                including all the data that is available.
        :param kwargs:          Additional keyword arguments.

        :return:                Clip with the processing applied to the selected frames.
        """

        frame_groups = self._frame_grouper.group_frames_into_ranges(clip, wobbly_parsed.interlaced_fades, 5)

        # TODO: Add fade detection and adaptive selection methods. For now, just use Average.

        return replace_ranges(clip, FixInterlacedFades.Average(clip), frame_groups)

    @property
    def position(self) -> FilteringPositionEnum:
        """When to perform filtering."""

        return FilteringPositionEnum.PRE_DECIMATE



from vstools import CustomValueError

__all__ = [
    'WobblyParseError',
    'WobblyValidationError',
    'WobblyAttributeError',
    'NotAWobblyFileError',
]


class WobblyParseError(CustomValueError):
    """Raised when parsing wobbly files fails."""

    ...


class NotAWobblyFileError(WobblyParseError):
    """Raised when a file is not a wobbly file."""

    ...


class WobblyValidationError(WobblyParseError):
    """Raised when validation fails."""

    ...


class WobblyAttributeError(WobblyParseError, AttributeError):
    """Raised when an attribute is not found."""

    ...



import logging

from vsdeinterlace import QTempGaussMC
from vstools import (
    CustomValueError,
    DependencyNotFoundError,
    FieldBased,
    core,
    depth,
    expect_bits,
    get_prop,
    replace_ranges,
    vs,
)

from ...components.orphans import OrphanFrame, OrphanFrames
from ...data.parse import WobblyParser
from ...types import FilteringPositionEnum
from .abstract import AbstractProcessingStrategy

__all__ = [
    'MatchBasedOrphanQTGMCStrategy',
]

logger = logging.getLogger(__name__)


FieldMatchGroupT = list[int]


class _OrphanFieldSplitter:
    """Helper class that splits orphaned fields into separate lists based on their field match."""

    def split_fields(
        self, wobbly_parsed: WobblyParser
    ) -> tuple[FieldMatchGroupT, FieldMatchGroupT, FieldMatchGroupT, FieldMatchGroupT]:
        orphan_n, orphan_b, orphan_u, orphan_p = [], [], [], []

        for frame in wobbly_parsed.orphan_frames:
            match frame.match:
                case 'b':
                    orphan_b.append(frame.frame)
                case 'n':
                    orphan_n.append(frame.frame)
                case 'u':
                    orphan_u.append(frame.frame)
                case 'p':
                    orphan_p.append(frame.frame)
                case _:
                    raise CustomValueError(f'Unknown field match: {frame.match} ({frame.frame})', self.split_fields)

        return orphan_n, orphan_b, orphan_u, orphan_p


class MatchBasedOrphanQTGMCStrategy(AbstractProcessingStrategy):
    """Strategy for dealing with orphan fields using match-based deinterlacing."""

    def __init__(self, thr: float = 0.0025, qtgmc_obj: QTempGaussMC | None = None) -> None:
        """
        :param thr:             Threshold for deinterlacing orphan fields.

                                If the difference between an orphan field and its adjacent same-type field
                                (as determined by field match) exceeds this value, the orphan field will be deinterlaced.
                                This helps avoid unnecessary deinterlacing for minor changes like mouth flaps,
                                while ensuring that significant differences trigger deinterlacing as intended.

                                Lower values will deinterlace more, but may also deinterlace fields that don't really need it.
                                Valid values are between 0 and 1.

                                Default: 0.0025.
        :param qtgmc_obj:       The `vsdeinterlace.QTempGaussMC` object to use. If not provided, a new one will be created.
                                Every relevant method should be called on this object by the user,
                                up until the "deinterlace" method, which will be called in this strategy.

                                Default: Automatically created by the strategy.
        """

        if not 0 <= thr <= 1:
            raise CustomValueError(f'Threshold must be between 0 and 1, not {thr}!', self.__init__)

        if qtgmc_obj is not None and not isinstance(qtgmc_obj, QTempGaussMC):
            raise CustomValueError(
                f'QTGMC object must be an instance of vsdeinterlace.QTempGaussMC, not {type(qtgmc_obj)}!', self.__init__
            )

        self.thr = thr
        self.qtgmc_obj = qtgmc_obj
        self._match_grouper = _OrphanFieldSplitter()

    def apply(self, clip: vs.VideoNode, wobbly_parsed: WobblyParser) -> vs.VideoNode:
        """
        Apply match-based deinterlacing to the given frames using QTGMC.

        This works by using the field match applied to orphan fields
        to determine which field is the correct one to keep.
        The other field gets deinterlaced.

        :param clip:            The clip to process.
        :param wobbly_parsed:   The parsed wobbly file. See the `WobblyParser` class for more information,
                                including all the data that is available.

        :return:                Clip with the processing applied to the selected frames.
        """

        clip = clip.std.SetFrameProps(wobbly_orphan_deint=False)

        field_order = FieldBased.from_param_or_video(wobbly_parsed.field_order, clip)

        clip = field_order.apply(clip)
        clip, orphans = self._should_deinterlace(clip, wobbly_parsed)

        if not orphans:
            return clip

        clip, bits = expect_bits(clip, 16)

        if self.qtgmc_obj is None:
            self.qtgmc_obj = self._qtgmc(clip)
        else:
            self.qtgmc_obj.clip = clip

        deint = self.qtgmc_obj.deinterlace()  # type: ignore

        assert isinstance(deint, vs.VideoNode)

        deint = deint[field_order.is_tff :: 2]
        deint = deint.std.SetFrameProps(wobbly_orphan_deint=True)

        frames = [o.frame for o in orphans if o.match in ('b', 'n')]

        logging.debug(f'Deinterlaced {len(frames)} orphan fields: {frames}')

        out = replace_ranges(clip, deint, frames)

        return depth(out, bits)

    @property
    def position(self) -> FilteringPositionEnum:
        """When to perform filtering."""

        return FilteringPositionEnum.PRE_DECIMATE

    def _should_deinterlace(self, clip: vs.VideoNode, wobbly_parsed: WobblyParser) -> [vs.VideoNode, OrphanFrames]:
        """
        Determine if the clip should be deinterlaced.

        We do this by splitting the fields and comparing the similarity between them.
        Which fields get compared depends on the field order.

        :param clip: The clip to process.

        :return: A clip with the fieldmatches applied if it shouldn't be deinterlaced, and a list of orphans to deinterlace.
        """

        if not (orphan_fields := wobbly_parsed.orphan_frames):
            return clip, OrphanFrames([])

        if not hasattr(core, 'vszip'):
            raise DependencyNotFoundError(self.apply, 'vszip')

        is_tff = wobbly_parsed.field_order.is_tff

        orphans_to_keep = list[OrphanFrame]()
        orphans_to_deint = list[OrphanFrame]()

        for match in ('b', 'n'):
            for orphan in orphan_fields.find_matches(match):
                offset = -1 if match == 'b' else 1 if match == 'n' else 0

                sep_ref = clip[orphan.frame + offset].std.SeparateFields()[is_tff]
                sep_curr = clip[orphan.frame].std.SeparateFields()[is_tff]

                sep_avg = get_prop(sep_ref.vszip.PlaneAverage([0], sep_curr), 'psmDiff', float)

                logging.debug(
                    f'{orphan.frame} ({match}): {(sep_avg >= self.thr)=} '
                    f'(sep_ref={orphan.frame + offset}, sep_curr={orphan.frame}) ({sep_avg=})'
                )

                if sep_avg >= self.thr:
                    orphans_to_deint.append(orphan)
                else:
                    orphans_to_keep.append(orphan)

        orphans_to_deint = OrphanFrames(orphans_to_deint)

        if orphans_to_keep:
            clip = self._revert_field_matches(clip, wobbly_parsed, orphans_to_keep)

        return clip, orphans_to_deint

    def _revert_field_matches(
        self, clip: vs.VideoNode, wobbly_parsed: WobblyParser, orphans: OrphanFrames
    ) -> vs.VideoNode:
        reverted_matches = list(wobbly_parsed.field_matches.fieldhint_string)

        for orphan in orphans:
            reverted_matches[orphan.frame] = orphan.match

        reverted_matches = ''.join(reverted_matches)

        fh = clip.fh.FieldHint(None, wobbly_parsed.field_order.is_tff, reverted_matches)
        return replace_ranges(clip, fh.std.SetFrameProps(wobbly_orphan_deint=-1), [o.frame for o in orphans])

    def _qtgmc(self, clip: vs.VideoNode) -> QTempGaussMC:
        """Create a QTGMC object for the given clip."""

        tr = 1

        return (
            QTempGaussMC(clip)
            .prefilter(tr=tr)
            .analyze(blksize=128, refine=5)
            .denoise(tr=tr)
            .basic(tr=tr, thsad=480)
            .source_match(tr=tr, mode=QTempGaussMC.SourceMatchMode.TWICE_REFINED)
            .lossless(mode=QTempGaussMC.LosslessMode.PRESHARPEN)
            .sharpen(mode=QTempGaussMC.SharpMode.NONE)
            .back_blend()
            .sharpen_limit(mode=QTempGaussMC.SharpLimitMode.NONE)
            .final(tr=tr)
            .motion_blur()
        )



from typing import Any

from vstools import CustomTypeError, CustomValueError, FuncExceptT, SupportsString

__all__ = [
    'InvalidMatchError',
]


class InvalidMatchError(CustomValueError):
    """Raised when an invalid match character is used."""

    def __init__(
        self,
        func: FuncExceptT,
        invalid_matches: str | list[str],
        message: SupportsString = 'Invalid match(es)!',
        reason: Any = '{invalid_matches}',
        **kwargs: Any,
    ) -> None:
        super().__init__(message, func, reason, **kwargs, invalid_matches=invalid_matches)

    @classmethod
    def check(cls, func: FuncExceptT, matches: str | list[str]) -> None:
        """Check if the match is valid."""

        from ..components.matches import ValidMatchT

        try:
            matches = ''.join({m for m in matches})
        except TypeError:
            raise CustomTypeError('Matches must be a string or list of strings!', func, reason=type(matches))

        if invalid_matches := [m for m in matches if m not in set(ValidMatchT.__args__)]:
            raise cls(func, invalid_matches)



from dataclasses import dataclass, field
from typing import Self

from vstools import FieldBased, SPathLike, vs

from vswobbly.data.parse import WobblyParser
from vswobbly.types import FilteringPositionEnum

from .strategies.abstract import AbstractProcessingStrategy
from .strategies.base import ProcessingStrategyManager

__all__ = ['WobblyProcessor']


@dataclass
class WobblyProcessor(ProcessingStrategyManager):
    """Class for processing a videonode using the data parsed from a wobbly file."""

    work_clip: vs.VideoNode
    """The videonode to process."""

    parser: WobblyParser
    """The parsed wobbly data."""

    # Strategies for handling certain types of issues.
    strategies: list[AbstractProcessingStrategy] = field(default_factory=list)
    """
    The strategies to use during specific processing steps.
    This can be strategies like combed-frame handling, how to deinterlace orphan fields, etc.
    See the `AbstractProcessingStrategy` class for more information.
    """

    def __init__(
        self,
        parser: WobblyParser,
        work_clip: vs.VideoNode | None = None,
        strategies: list[AbstractProcessingStrategy] = [],
    ) -> None:
        if work_clip is None:
            work_clip = parser.work_clip

        self.work_clip = work_clip
        self.parser = parser
        self.strategies = strategies

    def __post_init__(self) -> None:
        if not isinstance(self.strategies, list):
            self.strategies = [self.strategies]

    @classmethod
    def from_file(
        cls,
        wobbly_filepath: SPathLike,
        strategies: list[AbstractProcessingStrategy] | None = None,
    ) -> Self:
        """Create a processor from a wobbly file."""

        return cls(
            WobblyParser.from_file(wobbly_filepath),
            strategies=strategies,
        )

    def apply(self, clip: vs.VideoNode | None = None) -> vs.VideoNode:
        """Apply the wobbly processing to the given clip."""

        self._init_process(clip)

        self.apply_post_source()
        self.apply_post_field_match()
        self.apply_pre_decimation()
        self.apply_post_decimation()

        return self.proc_clip

    def _init_process(self, clip: vs.VideoNode | None = None) -> None:
        """Initialize the process."""

        self.proc_clip = clip or self.parser.work_clip

        if self.parser.video_data.trim:
            self.proc_clip = self.proc_clip.std.Trim(self.parser.video_data.trim[0], self.parser.video_data.trim[1])

        self.init_strategies(self.parser, self.strategies)

    def apply_post_source(self) -> None:
        """Post-source filtering, followed by field matching."""

        self.apply_strategies_of_position(FilteringPositionEnum.POST_SOURCE)

        # This must be run here to ensure the matches are set to 'c' correctly prior to deinterlacing.
        if any('orphan' in str(strategy).lower() for strategy in (self.strategies or [])):
            self.parser.field_matches.set_orphans_to_combed_matches(self.parser.orphan_frames)

        self.parser.sections.set_patterns(self.parser.field_matches)
        self.proc_clip = self.parser.sections.set_props(self.proc_clip, wobbly_parsed=self.parser)
        self.proc_clip = self.parser.combed_frames.set_props(self.proc_clip)
        self.proc_clip = self.parser.interlaced_fades.set_props(self.proc_clip)
        self.proc_clip = self.parser.orphan_frames.set_props(self.proc_clip)
        self.proc_clip = self.parser.field_matches.apply(self.proc_clip)

    def apply_post_field_match(self) -> None:
        """Post-field matching filtering."""

        self.apply_strategies_of_position(FilteringPositionEnum.POST_FIELD_MATCH)
        self.proc_clip = self.parser.freeze_frames.apply(self.proc_clip)

    def apply_pre_decimation(self) -> None:
        """
        Pre-decimation filtering. This should explicitly happen *after* post-field matching filtering
        because that's how presets work in wobbly. This is followed by decimation.
        """

        self.apply_strategies_of_position(FilteringPositionEnum.PRE_DECIMATE)
        self.proc_clip = self.parser.decimations.apply(self.proc_clip)

    def apply_post_decimation(self) -> None:
        """Post-decimation filtering."""

        self.apply_strategies_of_position(FilteringPositionEnum.POST_DECIMATE)
        self.proc_clip = FieldBased.PROGRESSIVE.apply(self.proc_clip)



from dataclasses import dataclass
from typing import Any

from vstools import CustomValueError, SPath, VSFunctionNoArgs, core, vs

__all__ = [
    'WobblyVideo',
]


@dataclass
class WobblyVideo:
    """Class for holding wobbly video data."""

    src_file: SPath
    """The path to the source file."""

    work_clip: vs.VideoNode
    """The clip to work on. Trimmed if necessary."""

    trim: tuple[int, int] | None
    """The trim to apply to the source clip. Inclusive/inclusive."""

    source_filter: VSFunctionNoArgs
    """The source filter to index the source file."""

    def __init__(self, src_file: SPath, wob_data: dict[str, Any]) -> None:
        self.src_file = SPath(src_file)
        self._set_trim(wob_data)

        self.source_filter = self._validate_source_filter(wob_data.get('source filter', ''))
        self._set_source_clip(wob_data.get('input file', str(self.src_file).removesuffix('.wob')))

    def _set_trim(self, wob_data: dict[str, Any]) -> None:
        trim_data = wob_data.get('trim')

        if not trim_data:
            self.trim = None

            return

        self.trim = tuple(trim_data[0])

    def _validate_source_filter(self, source_filter: str | VSFunctionNoArgs) -> VSFunctionNoArgs:
        """Validate the source filter and return the filter function."""

        if not source_filter:
            raise CustomValueError('Source filter cannot be empty!', self)

        if hasattr(source_filter, '__call__'):
            return source_filter

        if not isinstance(source_filter, str):
            raise CustomValueError('Invalid source filter!', self, source_filter)

        source_filter = source_filter.strip()

        if len(parts := source_filter.split('.')) != 2:
            raise CustomValueError('Invalid source filter format!', self, source_filter)

        namespace, filter_name = parts

        if not hasattr(core, namespace):
            raise CustomValueError(f"Namespace '{namespace}' not found in core", self)

        namespace_obj = getattr(core, namespace)

        if not hasattr(namespace_obj, filter_name):
            raise CustomValueError(f"Function '{filter_name}' not found in {namespace}", self)

        return getattr(namespace_obj, filter_name)

    def _set_source_clip(self, src_file: SPath) -> None:
        """Index and set the source clip using the source filter, and trim if necessary."""

        try:
            self.work_clip = self.source_filter(src_file)
        except Exception as e:
            raise CustomValueError(f'Error indexing source clip: {e}', self.__class__)



from typing import Any, Iterable

from vstools import CustomTypeError, CustomValueError, FuncExceptT, SupportsString

__all__ = [
    'NegativeFrameError',
]


class NegativeFrameError(CustomValueError):
    """Exception for negative frame numbers."""

    def __init__(
        self,
        func: FuncExceptT,
        invalid_frames: int | list[int],
        message: SupportsString = 'Frame number(s) cannot be negative!',
        reason: Any = '{invalid_frame}',
        **kwargs: Any,
    ) -> None:
        super().__init__(message, func, reason, **kwargs, invalid_frame=invalid_frames)

    @classmethod
    def check(cls, func: FuncExceptT, frames: int | Iterable[int]) -> None:
        """Check if the frame number is negative."""

        if not frames:
            return

        if isinstance(frames, int):
            if frames < 0:
                raise cls(func, frames)

            return

        if not isinstance(frames, Iterable):
            raise CustomTypeError('Frame number must be an integer!', func, invalid_frame=frames, reason=type(frames))

        if any(not isinstance(frame, int) for frame in frames):
            raise CustomTypeError('Frame numbers must be integers!', func, reason=type(frames))

        for i, frame in enumerate(frames):
            if frame >= 0:
                if i > 0:  # Stop checking once we find a positive frame
                    raise cls(func, frames[:i])
                return

        raise cls(func, frames)



# ruff: noqa: F401, F403

from .combed import *
from .custom_lists import *
from .decimations import *
from .freeze import *
from .ifades import *
from .matches import *
from .orphans import *
from .presets import *
from .sections import *
from .types import *
from .video import *



from typing import Any

from ..exceptions import WobblyValidationError


class WobblyValidator:
    """Class for validating wobbly files."""

    @staticmethod
    def validate_json_structure(data: dict[str, Any]) -> None:
        """Validate the JSON structure of a wobbly file."""

        required_fields = {
            'input file',
            'source filter',
        }

        missing = required_fields - set(data.keys())

        if missing:
            raise WobblyValidationError(f'Missing required fields: {missing}', WobblyValidator.validate_json_structure)

    @staticmethod
    def validate_version(data: dict[str, Any]) -> None:
        """Validate the version of a wobbly file."""

        if (version := data.get('wobbly version', 0)) < 6:
            raise WobblyValidationError(
                f'Unsupported wobbly version: {version}. Minimum supported version is 8. Please update!',
                WobblyValidator.validate_version,
            )



from stgpytools import CustomStrEnum

__all__ = [
    'FilteringPositionEnum',
]


class FilteringPositionEnum(CustomStrEnum):
    """Enum denoting when to perform filtering."""

    POST_SOURCE = 'post source'
    """Perform filtering on the source clip."""

    POST_FIELD_MATCH = 'post field match'
    """Perform filtering after field matching."""

    PRE_DECIMATE = 'pre decimate'
    """Perform filtering before decimation."""

    POST_DECIMATE = 'post decimate'
    """Perform filtering after decimation."""



"""Wobbly Vapoursynth Functions."""

__version__ = '0.2.2'

__author__ = 'LightArrowsEXE <LightArrowsReboot@gmail.com>'
__maintainer__ = __author__

__author_name__, __author_email__ = [x[:-1] for x in __author__.split('<')]
__maintainer_name__, __maintainer_email__ = [x[:-1] for x in __maintainer__.split('<')]

if __name__ == '__github__':
    print(__version__)



from dataclasses import dataclass
from typing import Any

from vstools import CustomValueError, VSFunctionNoArgs, fallback, replace_ranges, vs

from ..exceptions import NegativeFrameError

__all__ = [
    'InterlacedFade',
    'InterlacedFades',
]


@dataclass
class InterlacedFade:
    """Class for holding an interlaced fade."""

    frame: int
    """The frame number."""

    field_difference: float
    """The field difference."""

    def __post_init__(self):
        NegativeFrameError.check(self.__class__, self.frame)

        if not 0 <= self.field_difference <= 1:
            raise CustomValueError('Field difference must be between 0 and 1!', self, self.field_difference)

    def apply(self, clip: vs.VideoNode, filter: VSFunctionNoArgs | None = None) -> vs.VideoNode:
        """
        Apply the interlaced fade to the frame of the given clip using the specified filter.

        If no filter is provided, it will use `vsdeinterlace.fix_interlaced_fades`.

        :param clip:        The clip to process.
        :param filter:      The filter to use. The callable must only accept a clip as an argument.

        :return:            The clip with the interlaced fade applied.
        """

        if filter is None:
            from vsdeinterlace import fix_interlaced_fades

            filter = fix_interlaced_fades

        if not callable(filter):
            raise CustomValueError('Filter must be a callable!', self, filter)

        return filter(clip)


class InterlacedFades(list[InterlacedFade]):
    """Class for holding a list of interlaced fades."""

    def __init__(self, fades: list[InterlacedFade] | list[dict[str, Any]] | None = None, **kwargs) -> None:
        if isinstance(fades, list) and fades and isinstance(fades[0], dict):
            fades = [InterlacedFade(**fade) for fade in fades]

        super().__init__(fades or [])

        NegativeFrameError.check(self.__class__, [fade.frame for fade in fallback(self, [])])

    def apply(self, clip: vs.VideoNode, filter: VSFunctionNoArgs | None = None) -> vs.VideoNode:
        """
        Apply the interlaced fades to the clip using the specified filter.

        This filter is applied on all frames, so to properly handle dark/bright
        and other similar fades, you must handle them in the given filter.

        If no filter is provided, it will use `vsdeinterlace.fix_interlaced_fades`.

        :param clip:        The clip to process.
        :param filter:      The filter to use. The callable must only accept a clip as an argument.

        :return:            The clip with the interlaced fade applied.
        """

        if not self:
            return clip

        if filter is None:
            from vsdeinterlace import fix_interlaced_fades

            filter = fix_interlaced_fades

        if not callable(filter):
            raise CustomValueError('Filter must be a callable!', self, filter)

        return replace_ranges(clip, filter(clip), lambda n: n in set().union(*(fade.frame for fade in self)))

    def set_props(self, clip: vs.VideoNode) -> vs.VideoNode:
        """Set the interlaced fade properties on the clip."""

        return replace_ranges(
            clip.std.SetFrameProps(WobblyInterlacedFades=False),
            clip.std.SetFrameProps(WobblyInterlacedFades=True),
            [fade.frame for fade in self],
        )

    @classmethod
    def wob_json_key(cls) -> str:
        """The JSON key for interlaced fades."""

        return 'interlaced fades'



from dataclasses import dataclass, field
from typing import Any, Self

from vstools import FieldBased, FieldBasedT, SPath, SPathLike, vs

from ..components import (
    CombedFrames,
    CustomLists,
    Decimations,
    FieldMatches,
    FreezeFrames,
    InterlacedFades,
    OrphanFrames,
    Presets,
    Sections,
    WobblyVideo,
)

__all__ = [
    'WobblyParser',
]


@dataclass
class WobblyParser:
    """Class for parsing wobbly files."""

    file_path: SPath
    """The path to the wobbly file."""

    work_clip: vs.VideoNode
    """The clip to work on."""

    video_data: WobblyVideo
    """Source clip information."""

    field_order: FieldBasedT = FieldBased.TFF
    """Field order of the source clip."""

    sections: Sections = field(default_factory=Sections)
    """Sections of the wobbly file."""

    field_matches: FieldMatches = field(default_factory=FieldMatches)
    """List of field matches."""

    decimations: Decimations = field(default_factory=Decimations)
    """List of frames to decimate."""

    presets: Presets = field(default_factory=Presets)
    """List of filtering presets."""

    custom_lists: CustomLists = field(default_factory=CustomLists)
    """List of custom filtering ranges."""

    freeze_frames: FreezeFrames = field(default_factory=FreezeFrames)
    """List of freeze frames."""

    interlaced_fades: InterlacedFades = field(default_factory=InterlacedFades)
    """List of interlaced fades."""

    combed_frames: CombedFrames = field(default_factory=CombedFrames)
    """List of combed frames."""

    orphan_frames: OrphanFrames = field(default_factory=OrphanFrames)
    """List of orphan frames."""

    def __init__(
        self,
        file_path: SPath,
        work_clip: vs.VideoNode,
        video_data: WobblyVideo,
        field_order: FieldBasedT,
        sections: Sections | None = None,
        field_matches: FieldMatches | None = None,
        decimations: Decimations | None = None,
        presets: Presets | None = None,
        custom_lists: CustomLists | None = None,
        freeze_frames: FreezeFrames | None = None,
        interlaced_fades: InterlacedFades | None = None,
        combed_frames: CombedFrames | None = None,
        orphan_frames: OrphanFrames | None = None,
    ) -> None:
        self.file_path = file_path
        self.work_clip = work_clip
        self.video_data = video_data
        self.field_order = field_order

        self.sections = sections or Sections()
        self.field_matches = field_matches or FieldMatches()
        self.decimations = decimations or Decimations()
        self.presets = presets or Presets()
        self.custom_lists = custom_lists or CustomLists()
        self.freeze_frames = freeze_frames or FreezeFrames()
        self.interlaced_fades = interlaced_fades or InterlacedFades()
        self.combed_frames = combed_frames or CombedFrames()
        self.orphan_frames = orphan_frames or OrphanFrames()

    @classmethod
    def from_file(cls, file_path: SPathLike) -> Self:
        """Parse a wobbly object from a wobbly file."""

        from .builder import WobblyBuilder

        return WobblyBuilder(file_path).build()

    @staticmethod
    def _get_video_data(wob_file: SPath, data: dict[str, Any]) -> WobblyVideo:
        """Get the video data."""

        return WobblyVideo(wob_file.as_posix(), data.get('trim', None), data.get('source filter', ''))

    @staticmethod
    def _get_fieldbased_data(data: dict[str, Any]) -> FieldBasedT:
        """Get the fieldbased data."""

        vivtc_params = data.get('vfm parameters', {})
        order = bool(vivtc_params.get('order', 1))

        return FieldBased.from_param(order)



from dataclasses import dataclass
from typing import Self

from vstools import CustomValueError, fallback, replace_ranges, vs

from ..exceptions import NegativeFrameError
from .matches import FieldMatches, ValidMatchT

__all__ = [
    'OrphanFrame',
    'OrphanFrames',
]


@dataclass
class OrphanFrame:
    """Class for holding an orphan frame."""

    frame: int
    """The frame number."""

    match: ValidMatchT
    """
    The match used for this frame.
    Used to determine how to handle the frame by the orphan handler.
    """

    def __post_init__(self) -> None:
        NegativeFrameError.check(self.__class__, self.frame)

        if self.match not in ValidMatchT.__args__:  # type: ignore
            raise CustomValueError(f'Invalid match character: {self.match}', self)

        if self.match == 'c':
            raise CustomValueError("Orphan frames cannot have a 'c' match!", self)

    def __str__(self) -> str:
        return f'Frame {self.frame} ({self.match})'


class OrphanFrames(list[OrphanFrame]):
    """Class for holding orphan frames."""

    def __init__(self, frames: list[OrphanFrame] | None = None) -> None:
        super().__init__(frames or [])

        self.sort(key=lambda x: x.frame)

        NegativeFrameError.check(self.__class__, [frame.frame for frame in fallback(self, [])])

        seen = set()
        unique_frames = []

        for frame in self:
            if frame.frame not in seen:
                seen.add(frame.frame)
                unique_frames.append(frame)

        self[:] = unique_frames

    def __init_subclass__(cls) -> None:
        """Create matches properties for the class."""

        for match in cls.ValidMatchT.__args__:  # type: ignore
            if match == 'c':
                continue

            def create_match_property(match: cls.ValidMatchT) -> property:
                """Create a match property for the class."""

                def getter(self: OrphanFrames) -> list[OrphanFrame]:
                    """Get all frames with a specific match."""

                    return self.find_matches(match)

                return property(getter, doc=f"Get all frames with a '{match}' match.")

            setattr(cls, f'{match}_matches', create_match_property(match))

    def __str__(self) -> str:
        return ', '.join(str(frame) for frame in self)

    @classmethod
    def from_sections(cls, sections: 'Sections', matches: 'FieldMatches') -> Self:  # noqa: F821
        """Create orphan frames from sections and matches."""

        orphans = []

        for idx, section in enumerate(sections):
            if matches[section.start] == 'n':
                orphans.append(OrphanFrame(section.start, 'n'))

            end_frame = sections[idx + 1].start - 1 if idx < len(sections) - 1 else len(matches) - 1

            if matches[end_frame] == 'b':
                orphans.append(OrphanFrame(end_frame, 'b'))

        return OrphanFrames(orphans)

    def find_frame(self, frame: int) -> OrphanFrame | None:
        """Find a frame in the list."""

        try:
            return next(f for f in self if f.frame == frame)
        except StopIteration:
            return None

    def find_matches(self, match: ValidMatchT) -> Self:
        """Find all frames with a specific match."""

        return OrphanFrames([f for f in self if f.match == match])

    def set_props(self, clip: vs.VideoNode) -> vs.VideoNode:
        """Set the orphan frame properties on the clip."""

        if not self:
            return clip

        for match in ValidMatchT.__args__:  # type: ignore
            if match == 'c':
                continue

            clip = replace_ranges(
                clip.std.SetFrameProps(WobblyOrphanFrame=False),
                clip.std.SetFrameProps(WobblyOrphanFrame=match),
                self.find_matches(match).frames,
            )

        return clip

    @property
    def frames(self) -> list[int]:
        """Get all frames in the list."""

        return [frame.frame for frame in self]



import json
from dataclasses import dataclass
from typing import Any

from vstools import FieldBased, FieldBasedT, FileNotExistsError, FileWasNotFoundError, SPath, SPathLike

from ..components import (
    CombedFrames,
    CustomList,
    CustomLists,
    Decimations,
    FieldMatches,
    FreezeFrame,
    FreezeFrames,
    InterlacedFade,
    InterlacedFades,
    Preset,
    Presets,
    Section,
    Sections,
    WobblyVideo,
)
from ..data.parse import WobblyParser
from ..data.validation import WobblyValidator
from ..exceptions import NotAWobblyFileError, WobblyParseError
from ..types import FilteringPositionEnum
from ..util import to_snake_case

__all__ = ['WobblyBuilder']


@dataclass
class WobblyBuilder:
    """Builder class for constructing WobblyParser instances."""

    file_path: SPath
    _data: dict[str, Any] | None = None

    def __init__(self, file_path: SPathLike) -> None:
        self.file_path = SPath(file_path)

    def build(self) -> WobblyParser:
        """Build a WobblyParser instance."""

        self._check_file_path()
        self._load_data()

        WobblyValidator.validate_version(self._data)
        WobblyValidator.validate_json_structure(self._data)

        video_data = self._build_video_data()
        field_order = self._build_field_order()
        video_data.work_clip = field_order.apply(video_data.work_clip)

        return WobblyParser(
            file_path=SPath(self.file_path),
            work_clip=video_data.work_clip,
            video_data=video_data,
            field_order=field_order,
            **self._parse_data(),
        )

    def _check_file_path(self) -> None:
        """Check if the file path is valid."""

        if not self.file_path.exists():
            raise FileWasNotFoundError(f"File path does not exist: '{self.file_path}'", self)

        if not self.file_path.is_file():
            raise FileNotExistsError(f"File path is not a file: '{self.file_path}'", self)

        if not self.file_path.suffix == '.wob':
            raise NotAWobblyFileError('You must provide a wobbly file!', self)

        # We check for 2 bytes because wibbly writes 2 bytes when initializing
        if self.file_path.get_size() <= 2:
            raise WobblyParseError(f"File is empty: '{self.file_path}'", self)

    def _load_data(self) -> None:
        """Load the wobbly data."""

        with open(self.file_path, 'r') as file:
            self._data = json.load(file)

    def _build_video_data(self) -> WobblyVideo:
        return WobblyVideo(SPath(self.file_path).as_posix(), self._data)

    def _build_field_order(self) -> FieldBasedT:
        vivtc_params = self._data.get('vfm parameters', {})
        order = bool(vivtc_params.get('order', 1))

        return FieldBased.from_param(order)

    def _parse_data(self) -> dict[str, Any]:
        """Parse the wobbly data into their respective component classes."""

        parsed_data = self._parse_components(
            {
                Presets.wob_json_key(): ('presets', Presets, Preset),
                FieldMatches.wob_json_key(): ('field_matches', FieldMatches, str),
                CombedFrames.wob_json_key(): ('combed_frames', CombedFrames, int),
                Decimations.wob_json_key(): ('decimations', Decimations, int),
                Sections.wob_json_key(): ('sections', Sections, Section),
                InterlacedFades.wob_json_key(): ('interlaced_fades', InterlacedFades, InterlacedFade),
                CustomLists.wob_json_key(): ('custom_lists', CustomLists, CustomList),
                FreezeFrames.wob_json_key(): ('freeze_frames', FreezeFrames, FreezeFrame),
            }
        )

        self._build_orphan_frames(parsed_data)

        return parsed_data

    def _parse_components(self, component_map: dict[str, tuple[str, type, type]]) -> dict[str, Any]:
        """Parse each component from the wobbly data."""

        parsed_data = {}

        for wob_key, (attr_name, container_class, item_class) in component_map.items():
            if not self._should_parse_component(wob_key):
                continue

            data = self._data[wob_key]
            processed_items = self._get_processed_items(data, item_class)
            parsed_data[attr_name] = container_class(processed_items)

        return parsed_data

    def _should_parse_component(self, wob_key: str) -> bool:
        """Check if component should be parsed."""

        return wob_key in self._data and self._data[wob_key]

    def _get_processed_items(self, data: list, item_class: type) -> list:
        """Get processed items based on data type and item class."""

        if item_class == Section:
            return self._process_sections(data)

        if isinstance(data[0], dict):
            return self._process_dict_items(data, item_class)

        if isinstance(data[0], list):
            return self._process_list_items(data, item_class)

        return self._process_simple_items(data, item_class)

    def _process_sections(self, data: list[dict]) -> list[Section]:
        """Process sections and handle their presets."""

        preset_lookup = self._build_preset_lookup()
        processed_items = []

        for section in data:
            section_dict = self._to_snake_case(section)
            section_presets = []

            for preset_list in section_dict.get('presets', []):
                if isinstance(preset_list, str):
                    if preset_list in preset_lookup:
                        section_presets.append(preset_lookup[preset_list])
                elif isinstance(preset_list, list):
                    section_presets.extend(preset_lookup[p] for p in preset_list if p in preset_lookup)

            section_dict['presets'] = section_presets
            processed_items.append(Section(**section_dict))

            if not section_presets:
                continue

            self._add_section_presets_to_custom_lists(section_dict, section_presets, data)

        return processed_items

    def _build_preset_lookup(self) -> dict[str, Preset]:
        """Build lookup dictionary for presets."""

        all_presets = self._data.get('presets', [])

        return {
            p['name']: Preset(name=self._to_snake_case(p['name']), **{k: v for k, v in p.items() if k != 'name'})
            for p in all_presets
        }

    def _has_valid_presets(self, section_dict: dict) -> bool:
        """Check if section has valid presets."""

        return bool(section_dict.get('presets', []))

    def _add_section_presets_to_custom_lists(
        self, section_dict: dict, presets: list[Preset], all_sections: list[dict]
    ) -> None:
        """Add section presets to custom lists."""

        if 'custom lists' not in self._data:
            self._data['custom lists'] = []

        start = section_dict['start']
        section_idx = next(i for i, s in enumerate(all_sections) if s['start'] == start)
        end = self._get_section_end(section_idx, all_sections)

        for preset in presets:
            self._data['custom lists'].append(
                {
                    'name': f'section_{start}_{end}_{preset.name}',
                    'preset': preset,
                    'position': FilteringPositionEnum.PRE_DECIMATE,
                    'frames': [(start, end)],
                }
            )

    def _get_section_end(self, section_idx: int, all_sections: list[dict]) -> int:
        """Get end frame for section."""

        if section_idx < len(all_sections) - 1:
            return all_sections[section_idx + 1]['start'] - 1

        return len(self._data.get('field matches', []))

    def _process_dict_items(self, data: list[dict], item_class: type) -> list:
        """Process dictionary items into their respective dataclass instances."""

        if item_class == CustomList:
            preset_lookup = self._build_preset_lookup()

            return [
                item_class(
                    **{
                        **{to_snake_case(k): v for k, v in item.items() if k != 'preset' and k != 'frames'},
                        'preset': preset_lookup.get(
                            item['preset'].name if isinstance(item['preset'], Preset) else item['preset']
                        ),
                        'frames': [tuple(frame) for frame in item['frames']],
                    }
                )
                for item in data
            ]

        return [item_class(**{to_snake_case(k): v for k, v in item.items()}) for item in data]

    def _process_list_items(self, data: list[list], item_class: type) -> list:
        """Process list items into their respective dataclass instances."""

        if item_class == FreezeFrame:
            return [item_class(first=item[0], last=item[1], replacement=item[2]) for item in data]

        return [item_class(**{'frames': item}) for item in data]

    def _process_simple_items(self, data: list, item_class: type) -> list:
        """Process simple type items (int, str, etc.)."""

        return [item_class(x) for x in data]

    def _to_snake_case(self, item: dict[Any, Any] | str) -> dict[Any, Any] | str:
        """Convert dictionary keys from space-separated to snake_case."""

        if isinstance(item, str):
            return to_snake_case(item)

        return {
            to_snake_case(k) if isinstance(v, str) else k: to_snake_case(v) if isinstance(v, str) else v
            for k, v in item.items()
        }

    def _build_orphan_frames(self, parsed_data: dict) -> None:
        """Add orphan frames if both sections and field matches exist."""

        from ..components import OrphanFrames

        sections = parsed_data.get(Sections.wob_json_key())
        matches = parsed_data.get('field_matches')

        if not sections or not matches:
            return

        parsed_data['orphan_frames'] = OrphanFrames.from_sections(sections, matches)



from abc import ABC, abstractmethod
from typing import Any

from vstools import vs

from ...data.parse import WobblyParser
from ...types import FilteringPositionEnum

__all__ = [
    'AbstractProcessingStrategy',
]


class AbstractProcessingStrategy(ABC):
    """
    Abstract base class for processing strategies.

    To write your own strategy, inherit from this class and implement
    the `apply` method and the `position` property.

    :param kwargs:              Additional keyword arguments.
                                These will be passed to the strategy when it is applied.
    """

    def __init__(self, **kwargs: Any) -> None:
        self._kwargs = kwargs

    @abstractmethod
    def apply(self, clip: vs.VideoNode, wobbly_parsed: WobblyParser) -> vs.VideoNode:
        """
        Apply the strategy to the given frames.

        :param clip:                The clip to process.
        :param wobbly_parsed:       The parsed wobbly file. See the `WobblyParser` class for more information,
                                    including all the data that is available.

        :return:                    Clip with the processing applied to the selected frames.
        """

        # It's recommended to use `replace_ranges` to apply your strategy.
        # This is gonna be the fastest and most straightforward way to do it,
        # and allows you to easily access surrounding frames if necessary.
        # See the implementations in this library for examples.

    @property
    @abstractmethod
    def position(self) -> FilteringPositionEnum:
        """When to apply the strategy."""

        return FilteringPositionEnum.POST_DECIMATE



import ast
from dataclasses import dataclass
from typing import Any

from vstools import CustomValueError, vs

__all__ = [
    'Preset',
    'Presets',
]


@dataclass
class Preset:
    """Class for holding a preset."""

    name: str
    """The name of the preset."""

    contents: str
    """The contents of the preset."""

    def __post_init__(self) -> None:
        """Validate and check safety of preset after initialization."""

        if not self.name:
            raise CustomValueError('Preset name cannot be empty!', self._err_name)

        if not self.contents:
            del self
            return

        self._check_code_exec()

    def __call__(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        """Apply the preset code to a given clip."""

        return self.apply(clip, **kwargs)

    def __str__(self) -> str:
        return f'{self.name}:\n\n{self.contents}'

    def apply(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        """Apply the preset code to a given clip."""

        namespace = {'clip': clip}

        exec(self.contents, namespace)

        return namespace['clip']

    def _check_code_exec(self) -> None:
        """Check if the preset is executable."""

        try:
            compile(self.contents, '<string>', 'exec')
        except SyntaxError as e:
            raise CustomValueError(f'Invalid Python code in preset contents: {e}', self._err_name)
        except Exception as e:
            raise CustomValueError(f'Invalid preset contents: {e}', self._err_name)

        tree = ast.parse(self.contents)

        has_clip = False

        for node in ast.walk(tree):
            if isinstance(node, ast.Name) and node.id == 'clip':
                has_clip = True

                break

        if not has_clip:
            raise CustomValueError('Preset must use the "clip" variable', self._err_name)

    def _check_unsafe_node(self, node: ast.AST, depth: int = 0) -> None:
        """Check if an individual AST node contains unsafe operations."""

        if depth > 100:  # Add recursion depth limit
            raise CustomValueError('Maximum recursion depth exceeded while checking preset safety', self._err_name)

        if not isinstance(node, ast.AST):
            return

        try:
            if isinstance(node, ast.Call):
                if isinstance(node.func, ast.Name):
                    if node.func.id in {'eval', 'exec', 'open', 'system'}:
                        raise CustomValueError(f'Unsafe function call: {node.func.id}()', self._err_name)

            for child in ast.iter_child_nodes(node):  # Use iter_child_nodes instead of walk
                self._check_unsafe_node(child, depth + 1)

        except CustomValueError:
            raise
        except Exception as e:
            raise CustomValueError(f'Error checking preset safety: {e}', self._err_name)

    @property
    def _err_name(self) -> str:
        """The error name for the preset."""

        if not self.name:
            return self.__class__.__name__

        return f'{self.__class__.__name__}:{self.name}'


class Presets(list[Preset]):
    """Class for holding a list of presets."""

    def __init__(self, presets: list[Preset] = []) -> None:
        super().__init__(presets)

    def __str__(self) -> str:
        return ', '.join(self)

    @classmethod
    def wob_json_key(cls) -> str:
        """The JSON key for presets."""

        return 'presets'



from vstools import fallback, vs

from ..exceptions import NegativeFrameError
from ..util import deduplicate_list

__all__ = [
    'Decimations',
]


class Decimations(list[int]):
    """Class for holding a sorted list of decimated frame indices."""

    def __init__(self, decimations: list[int] | None = None) -> None:
        super().__init__(decimations or [])

        NegativeFrameError.check(self.__class__, fallback(self, []))

        self.sort()

        self[:] = deduplicate_list(self)

    @classmethod
    def wob_json_key(cls) -> str:
        """The JSON key for decimated frames."""

        return 'decimated frames'

    def find_decimation(self, frame: int) -> int | None:
        """Find a decimation in the list."""

        try:
            return self[self.index(frame)]
        except ValueError:
            return None

    def apply(self, clip: vs.VideoNode) -> vs.VideoNode:
        """Apply the decimations to the clip."""

        if len(self) == 0:
            return clip

        dec = clip.std.DeleteFrames(self)

        return dec



from typing import Any, Iterator, Literal, Protocol, Sequence

from vstools import vs

__all__ = [
    'ValidMatchT',
    'PresetProtocol',
    'SectionProtocol',
    'SectionsProtocol',
]


ValidMatchT = Literal['p', 'c', 'n', 'b', 'u']
"""Type alias for valid match characters.

p: Match to previous field
c: Match to current field
n: Match to next field
b: Match to previous field (matches from opposite parity field)
u: Match to next field (matches from opposite parity field)

See `http://avisynth.nl/index.php/TIVTC/TFM` for more information.
"""


class PresetProtocol(Protocol):
    """Protocol defining the interface for presets."""

    name: str

    def apply(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode: ...


class SectionProtocol(Protocol):
    """Protocol defining the interface for sections."""

    start: int
    presets: Sequence[PresetProtocol]


class SectionsProtocol(Protocol):
    """Protocol defining the interface for a collection of sections."""

    def __iter__(self) -> Iterator[SectionProtocol]: ...



"""
Utility functions used throughout this library and is not meant to be used by other packages.
"""

from typing import TypeVar

__all__ = [
    'deduplicate_list',
    'to_snake_case',
]


T = TypeVar('T')


def deduplicate_list(items: list[T]) -> list[T]:
    """Deduplicate a list."""

    if not len(items) > 1:
        return items

    return list(dict.fromkeys(items))


def to_snake_case(key: str) -> str:
    """Convert a key to snake_case."""

    return '_'.join(key.strip().split(' '))



from vstools import fallback, replace_ranges, vs

from ..exceptions import NegativeFrameError
from ..util import deduplicate_list

__all__ = [
    'CombedFrames',
]


class CombedFrames(list[int]):
    """Class for holding combed frames."""

    def __init__(self, frames: list[int] | None = None) -> None:
        super().__init__(frames or [])

        NegativeFrameError.check(self.__class__, fallback(self, []))

        self.sort()

        self[:] = deduplicate_list(self)

    @classmethod
    def wob_json_key(cls) -> str:
        """The JSON key for combed frames."""

        return 'combed frames'

    def find_frame(self, frame: int) -> int | None:
        """Find a frame in the list."""

        if frame in self:
            return next(f for f in self if f == frame)

        return None

    def get_frame(self, frame: int) -> int:
        """Get a frame from the list, and raise an error if not found."""

        if frame < 0:
            raise NegativeFrameError(self.__class__, frame)

        if (found := self.find_frame(frame)) is None:
            raise ValueError(f'Frame {frame} not found!')

        return found

    def set_props(self, clip: vs.VideoNode) -> vs.VideoNode:
        """Set the combed frame properties on the clip."""

        return replace_ranges(
            clip.std.SetFrameProps(WobblyCombed=False), clip.std.SetFrameProps(WobblyCombed=True), self
        )



from vsdeinterlace import vinverse
from vstools import replace_ranges, vs

from ...data.parse import WobblyParser
from ...types import FilteringPositionEnum
from .abstract import AbstractProcessingStrategy

__all__ = [
    'DecombVinverseStrategy',
]


class DecombVinverseStrategy(AbstractProcessingStrategy):
    """Decomb strategy using vinverse to get rid of combed frames."""

    def apply(self, clip: vs.VideoNode, wobbly_parsed: WobblyParser) -> vs.VideoNode:
        """
        Use vinverse to get rid of combing on the given frames.

        Note that vinverse may be destructive. Be cautious when applying this strategy.

        :param clip:            The clip to process.
        :param wobbly_parsed:   The parsed wobbly file. See the `WobblyParser` class for more information,
                                including all the data that is available.
        :param kwargs:          Additional keyword arguments.

        :return:                Clip with the processing applied to the selected frames.
        """

        return replace_ranges(clip, vinverse(clip), wobbly_parsed.combed_frames)

    @property
    def position(self) -> FilteringPositionEnum:
        """When to perform filtering."""

        return FilteringPositionEnum.PRE_DECIMATE



#!/usr/bin/env python3

from pathlib import Path

import setuptools  # type:ignore[import-untyped]

long_description = Path('README.md').read_text(encoding='utf-8')
install_requires = Path('requirements.txt').read_text(encoding='utf-8')

package_name = 'vswobbly'

exec(Path(f'{package_name}/_metadata.py').read_text(encoding='utf-8'), meta := dict[str, str]())

setuptools.setup(
    name=package_name,
    version=meta['__version__'],
    author=meta['__author_name__'],
    author_email=meta['__author_email__'],
    maintainer=meta['__maintainer_name__'],
    maintainer_email=meta['__maintainer_email__'],
    description=meta['__doc__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=[
        package_name,
        f'{package_name}.components',
        f'{package_name}.data',
        f'{package_name}.exceptions',
        f'{package_name}.process',
        f'{package_name}.process.strategies',
    ],
    package_data={
        package_name: ['py.typed', 'models/shaders/*/*.onnx'],
    },
    install_requires=install_requires,
    project_urls={
        'Source Code': 'https://github.com/Jaded-Encoding-Thaumaturgy/vs-wobbly',
        'Contact': 'https://discord.gg/XTpc6Fa9eB',
    },
    classifiers=[
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'Intended Audience :: Other Audience',
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Typing :: Typed',
        'Topic :: Multimedia :: Video',
        'Topic :: Multimedia :: Video :: Display',
    ],
    python_requires='>=3.12',
)



from typing import Any
from vstools import vs

from ...components import CustomList
from ...data.parse import WobblyParser
from ...types import FilteringPositionEnum
from .abstract import AbstractProcessingStrategy

__all__ = [
    'CustomListStrategy',
]


class CustomListStrategy(AbstractProcessingStrategy):
    """Default strategy that applies a custom list defined directly in the wobbly file."""

    def __init__(self, custom_list: CustomList, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._custom_list = custom_list

    def apply(self, clip: vs.VideoNode, wobbly_parsed: WobblyParser) -> vs.VideoNode:
        """
        Apply the custom list.

        :param clip:            The clip to process.
        :param wobbly_parsed:   The parsed wobbly file. See the `WobblyParser` class for more information,
                                including all the data that is available.

        :return:                Clip with the processing applied to the selected frames.
        """

        return self._custom_list.apply(clip, wobbly_parsed.decimations)

    @property
    def position(self) -> FilteringPositionEnum:
        """When to perform filtering."""

        return self._custom_list.position



from dataclasses import dataclass

from vstools import CustomValueError, fallback, replace_ranges, vs

from ..exceptions import NegativeFrameError

__all__ = [
    'FreezeFrame',
    'FreezeFrames',
]


@dataclass
class FreezeFrame:
    """Class for holding a freeze frame."""

    first: int
    """The first frame to freeze."""

    last: int
    """The last frame to freeze."""

    replacement: int
    """The frame to replace the frozen frames with."""

    def __post_init__(self) -> None:
        NegativeFrameError.check(self.__class__, [self.first, self.last, self.replacement])

        if self.first > self.last:
            raise CustomValueError(f'First frame ({self.first}) must start before the last frame ({self.last})!', self)

    def __iter__(self) -> tuple[int, int, int]:
        """Return a tuple of first, last and replacement frames for unpacking."""

        return self.first, self.last, self.replacement


class FreezeFrames(list[FreezeFrame]):
    """List of frozen frames."""

    def __post_init__(self) -> None:
        NegativeFrameError.check(self.__class__, [freeze.first for freeze in fallback(self, [])])
        NegativeFrameError.check(self.__class__, [freeze.last for freeze in fallback(self, [])])
        NegativeFrameError.check(self.__class__, [freeze.replacement for freeze in fallback(self, [])])

        wrong_ranges = []

        for start, end in zip(self.first, self.last):
            if start > end:
                wrong_ranges.append((start, end))

        if wrong_ranges:
            raise CustomValueError(f'First frame must start before the last frame! ({wrong_ranges})', self)

    def apply(self, clip: vs.VideoNode) -> vs.VideoNode:
        """Apply the freeze frames to the clip."""

        if len(self) == 0:
            return clip

        # I realise this is slower! See _better_apply for a faster WIP implementation.
        for freeze in self:
            try:
                frozen = clip.std.FreezeFrames(freeze.first, freeze.last, freeze.replacement)
                frozen = frozen.std.SetFrameProps(WobblyFreeze=[freeze.first, freeze.last, freeze.replacement])
            except vs.Error as e:
                raise CustomValueError(f'Failed to apply freeze frames ({freeze}): {e}', self) from e

            clip = replace_ranges(clip, frozen, [(freeze.first, freeze.last)])

        return clip

    # TODO: Finish implementing this
    def _better_apply(self, clip: vs.VideoNode) -> vs.VideoNode:
        firsts, lasts, replacements = [], [], []

        for freeze in self:
            firsts.append(freeze.first)
            lasts.append(freeze.last)
            replacements.append(freeze.replacement)

            clip = replace_ranges(
                clip,
                clip.std.SetFrameProps(WobblyFreeze=[freeze.first, freeze.last, freeze.replacement]),
                [(freeze.first, freeze.last)],
            )

        # TODO: The current issue is that props don't persist if the replacement isn't in the range of (first, last).
        return clip.std.FreezeFrames(firsts, lasts, replacements)

    @classmethod
    def wob_json_key(cls) -> str:
        """The JSON key for freeze frames."""

        return 'frozen frames'



from bisect import bisect_left, bisect_right
from dataclasses import dataclass
from typing import Any, Iterable, Self

from vstools import CustomRuntimeError, CustomValueError, FrameRangesN, replace_ranges, vs

from ..types import FilteringPositionEnum
from .decimations import Decimations
from .types import PresetProtocol, SectionsProtocol

__all__ = ['CustomList', 'CustomLists']


@dataclass
class CustomList:
    """Class for holding a custom list."""

    name: str
    """The name of the custom list."""

    preset: PresetProtocol
    """The preset used for this custom list."""

    position: FilteringPositionEnum
    """When to apply the preset."""

    frames: FrameRangesN
    """The frames to apply the preset to."""

    def __init__(self, **kwargs) -> None:
        if 'frames' in kwargs and isinstance(kwargs['frames'], list):
            kwargs['frames'] = self._frames_to_ranges(kwargs['frames'])

        self.name = kwargs.get('name', '')
        self.preset = kwargs.get('preset', '')
        self.position = FilteringPositionEnum(kwargs.get('position', 'pre decimation'))
        self.frames = kwargs.get('frames', [])

    def __post_init__(self) -> None:
        """Validate the custom list after initialization."""

        if not self.frames:
            raise CustomValueError('Custom list frames cannot be empty!', self)

    def __call__(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        """Apply the custom list to a given clip."""

        return self.apply(clip, **kwargs)

    @staticmethod
    def _frames_to_ranges(frames: list[list[int]]) -> FrameRangesN:
        """Convert the frames to a list of frame ranges."""

        ranges = []
        invalid = []

        for frame in frames:
            if not isinstance(frame, (Iterable, int)):
                invalid.append(frame)
                continue

            if isinstance(frame, int):
                ranges.append(frame)
                continue

            if not isinstance(frame, tuple):
                frame = tuple(frame)

            if len(frame) != 2:
                invalid.append(frame)
                continue

            if frame[0] == frame[1]:
                ranges.append(frame[0])
                continue

            if frame[0] > frame[1]:
                invalid.append(frame)
                continue

            ranges.append(frame)

        if invalid:
            raise CustomValueError(f'Invalid frame ranges in custom list: {invalid}', CustomList)

        return ranges

    def apply(self, clip: vs.VideoNode, decimations: Decimations, **kwargs: Any) -> vs.VideoNode:
        """
        Apply the custom list to a given clip.

        If this custom list's :attr:`position` is :attr:`FilteringPositionEnum.POST_DECIMATE`,
        then this accounts for decimated frames by adjusting frame ranges
        based on the number of decimations that occur before each range endpoint.

        :param clip:            The clip to apply the custom list to.
        :param decimations:     The decimations to account for.

        :return:                The clip with the custom list applied.
        """

        try:
            flt = self.preset.apply(clip, **kwargs)
        except Exception as e:
            raise CustomRuntimeError(
                f"Error applying preset of custom list '{self.name}': "
                f'Invalid Python code in preset contents.\nOriginal error: {e}',
                self.apply,
            )

        for _range in self._frames_to_ranges(self.frames):
            range_flt = flt.std.SetFrameProps(
                WobblyPreset=str(self.preset), WobblyPresetPosition=self.position.value, WobblyPresetFrames=_range
            )

            if self.position is FilteringPositionEnum.POST_DECIMATE:
                first, last = _range if isinstance(_range, tuple) else (_range, _range)
                first -= bisect_left(decimations, first)
                last -= bisect_right(decimations, last)
                if first > last:
                    # Wobbly-generated scripts fail at runtime in this case, but that's annoying
                    continue
                effective_range = first if first == last else (first, last)
            else:
                effective_range = _range

            if isinstance(_range, tuple) and _range[1] >= clip.num_frames:
                _range = (_range[0], clip.num_frames - 1)

            explained_range = _range if _range == effective_range else f'{_range}, after decimation: {effective_range}'

            try:
                clip = replace_ranges(clip, range_flt, effective_range)
            except vs.Error as e:
                if 'invalid last frame' in str(e):
                    clip = replace_ranges(clip, range_flt, (effective_range[0], clip.num_frames - 1))
                else:
                    raise CustomRuntimeError(
                        f"Error applying custom list '{self.name}' (range: {explained_range}): {e}", self.apply
                    )
            except Exception as e:
                raise CustomRuntimeError(
                    f"Error applying custom list '{self.name}' (range: {explained_range}): {e}", self.apply
                )

        return clip


class CustomLists(list[CustomList]):
    """Class for holding a list of custom lists."""

    def __str__(self) -> str:
        return ', '.join(
            f'{custom_list.name}: preset={custom_list.preset}, '
            f'position={custom_list.position.name}, frames={custom_list.frames}'
            for custom_list in self
        )

    @classmethod
    def from_sections(cls, sections: SectionsProtocol) -> Self:
        """Create custom lists from sections."""

        return cls(
            CustomList(
                f'section_{section.start}',
                section.presets,
                FilteringPositionEnum.PRE_DECIMATE,
                FrameRangesN(section.start),
            )
            for section in sections
        )

    def apply(self, clip: vs.VideoNode, decimations: Decimations) -> vs.VideoNode:
        """Apply all custom lists to a given clip."""

        for custom_list in self:
            clip = custom_list.apply(clip, decimations)

        return clip

    @classmethod
    def wob_json_key(cls) -> str:
        """The JSON key for custom lists."""

        return 'custom lists'



from vstools import CustomNotImplementedError, CustomValueError, vs

from ...data.parse import WobblyParser
from ...types import FilteringPositionEnum
from .abstract import AbstractProcessingStrategy
from .custom_lists import CustomListStrategy

__all__ = ['ProcessingStrategyManager']


class ProcessingStrategyManager:
    """Class for managing and executing processing strategies in a specific order."""

    def init_strategies(
        self, wobbly_parsed: WobblyParser, strategies: list[AbstractProcessingStrategy] | None = None
    ) -> None:
        """Initialize and validate the list of strategies."""

        all_strategies = []

        if strategies:
            all_strategies.extend(
                [
                    strategy() if not isinstance(strategy, AbstractProcessingStrategy) else strategy
                    for strategy in strategies
                ]
            )

        all_strategies.extend(
            [value for name, value in vars(self).items() if name.endswith('_strategy') and value is not None]
        )

        all_strategies.extend(map(CustomListStrategy, wobbly_parsed.custom_lists))

        self._strategies = all_strategies

        self._ensure_strategies_callable()

    def apply_strategies_of_position(self, position: FilteringPositionEnum) -> vs.VideoNode:
        """Apply all strategies of a given position."""

        if not hasattr(self, '_strategies'):
            self.init_strategies(self.parser)

        strategies = self._get_strategies_for_position(position)

        for strategy in strategies:
            self.proc_clip = strategy.apply(self.proc_clip, wobbly_parsed=self.parser)
            # self.proc_clip = self._add_strategy_as_frame_prop(strategy)  # TODO: Implement this

        return self.proc_clip

    def _add_strategy_as_frame_prop(self, strategy: AbstractProcessingStrategy) -> None:
        """Add a strategy as a frame property."""

        raise CustomNotImplementedError('This method is not implemented yet!', self._add_strategy_as_frame_prop)

    def _ensure_strategies_callable(self) -> None:
        """Ensure that all strategies are callable."""

        uncallable = []

        for strategy in self._strategies:
            if not hasattr(strategy, 'apply') or not callable(strategy.apply):
                uncallable.append(strategy)

        if uncallable:
            raise CustomValueError(f'The following strategies are not callable: {uncallable}', self.init_strategies)

    def _get_strategies_for_position(self, position: FilteringPositionEnum) -> list[AbstractProcessingStrategy]:
        """Get all strategies that should run at the given position."""

        return [strategy for strategy in self._strategies if strategy.position == position]



from bisect import bisect_left
from dataclasses import dataclass, field
from math import ceil
from typing import Literal

from vstools import Keyframes, vs

from ..exceptions import NegativeFrameError
from .decimations import Decimations
from .matches import FieldMatches
from .types import PresetProtocol, SectionProtocol

__all__ = [
    'Section',
    'Sections',
]


@dataclass
class Section(SectionProtocol):
    """Class for holding a section."""

    start: int
    """The start frame number."""

    presets: list[PresetProtocol] = field(default_factory=list)
    """The presets used for this section."""

    dominant_pattern: Literal[0, 1, 2, 3, 4, -1] = -1
    """The dominant pattern for this section. -1 means unknown."""

    def __post_init__(self):
        NegativeFrameError.check(self, self.start)

    def set_pattern(self, pattern: Literal[0, 1, 2, 3, 4]):
        """Set the dominant pattern for this section."""

        self.dominant_pattern = pattern


class Sections(list[Section]):
    """Class for holding sections."""

    def __init__(self, sections: list[Section]) -> None:
        super().__init__(sections or [])

    def __str__(self) -> str:
        if not self:
            return ''

        return ', '.join(str(section) for section in self)

    def to_keyframes(self, decimations: Decimations) -> Keyframes:
        """
        Convert the sections to keyframes.

        Accounts for decimated frames by adjusting section start frames
        based on the number of decimations that occur before each section start.

        :param decimations:     The decimations to account for.

        :return:                A keyframes object representing the section start frames adjusted for decimations.
        """

        if not self:
            return Keyframes([])

        if not decimations:
            return Keyframes([section.start for section in self])

        keyframes = [section.start - bisect_left(decimations, section.start) for section in self]

        return Keyframes(keyframes)

    @classmethod
    def wob_json_key(cls) -> str:
        """The JSON key for sections."""

        return 'sections'

    def set_props(self, clip: vs.VideoNode, wobbly_parsed: 'WobblyParser') -> vs.VideoNode:  # noqa: F821
        """Set the section properties on the clip."""

        wclip = wobbly_parsed.work_clip
        # Ideally we get the cycle from the vfm params, but wobbly is hardcoded to 5 anyway.
        cycle = 5

        framerates = [wclip.fps.numerator / cycle * i for i in range(cycle, 0, -1)]

        fps_clips = [
            clip.std.AssumeFPS(None, int(fps), wclip.fps.denominator).std.SetFrameProps(
                WobblyCycleFps=int(fps // 1000), _DurationNum=int(fps), _DurationDen=wclip.fps.denominator
            )
            for fps in framerates
        ]

        max_dec = max(wobbly_parsed.decimations) + 1

        split_decimations = [
            [j for j in range(i * cycle, min((i + 1) * cycle, max_dec)) if j in wobbly_parsed.decimations]
            for i in range(ceil(max_dec / cycle))
        ]

        n_split_decimations = len(split_decimations)

        indices = [
            0 if (sd_idx := ceil(n / cycle)) >= n_split_decimations else len(split_decimations[sd_idx])
            for n in range(clip.num_frames)
        ]

        # Set pattern for each frame based on which section it falls into
        pattern_props = []

        for n in range(clip.num_frames):
            section_idx = 0

            for i, _ in enumerate(self):
                if i < len(self) - 1 and n >= self[i + 1].start:
                    continue

                section_idx = i
                break

            pattern_props.append(self[section_idx].dominant_pattern)

        return clip.std.FrameEval(lambda n: fps_clips[indices[n]].std.SetFrameProps(WobblyPattern=pattern_props[n]))

    def set_patterns(self, matches: FieldMatches) -> None:
        """Set the dominant patterns for all sections based on the matches."""

        import warnings

        warnings.warn(
            'Sections.set_patterns: This method is not yet implemented and will be implemented in a future version.',
            DeprecationWarning,
        )

        return
